/*
 * build_cpu6280.c
 *
 * build the cpu6280 core opcode handlers
 */

/* $Id: build_cpu6280.c,v 1.29 2000/02/11 03:26:09 nyef Exp $ */

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>

#include "types.h"

FILE *outfile;

int handlers[0x100];
int cur_handler;

const char file_header[] = "/*\n * %s\n *\n * opcode handlers for cpu6280 core\n *\n * MACHINE GENERATED. DO NOT EDIT.\n */\n\n#include \"cpu6280int.h\"\n\n";

const char file_footer[] = "/* EOF */\n";

typedef void (*instr_emitter_t)(void);

void emit(const char *format, ...)
{
    va_list args;

    va_start(args, format);
    vfprintf(outfile, format, args);
    va_end(args);
}

void emit_file_header(char *filename)
{
    emit(file_header, filename);
}

void emit_file_footer(void)
{
    emit(file_footer);
}

void clear_handlers(void)
{
    int i;

    for (i = 0; i < 0x100; i++) {
	handlers[i] = 0;
    }

    cur_handler = 1;
}

void emit_decode_table(void)
{
    int i;

    emit("ophandler_t cpu6280_decode[0x100] = {\n");
    
    for (i = 0; i < 0x100; i++) {
	if ((i & 3) == 0) {
	    emit("   ");
	}
	emit(" cpu6280_op_%d,", handlers[i]);
	if ((i & 3) == 3) {
	    emit(" /* 0x%02x - 0x%02x */\n", i & 0xfc, i);
	}
    }

    emit("};\n\n");
}

void emit_procheader(u8 opcode)
{
    emit("void cpu6280_op_%d(struct cpu6280_context *context)\n", cur_handler);
    emit("{\n");
    emit("    /* 0x%02x */\n", opcode);
    handlers[opcode] = cur_handler;
}

void emit_proctrailer(void)
{
    emit("}\n\n");

    cur_handler++;
}

struct {
    u8 opcode;
    char *instruction;
    char *flag;
    int set;
} flag_instructions[] = {
    {0x18, "CLC", "FLAG_C", 0},
    {0x38, "SEC", "FLAG_C", 1},
    {0xd8, "CLD", "FLAG_D", 0},
    {0xf8, "SED", "FLAG_D", 1},
    {0x58, "CLI", "FLAG_I", 0},
    {0x78, "SEI", "FLAG_I", 1},
    {0xb8, "CLV", "FLAG_V", 0},
};

void emit_flag_instructions(void)
{
    int i;

    for (i = 0; i < (sizeof(flag_instructions) / sizeof(flag_instructions[0])); i++) {
	emit_procheader(flag_instructions[i].opcode);
	emit("    /* %s */\n", flag_instructions[i].instruction);
	if (flag_instructions[i].set) {
	    emit("    context->flags |= %s;\n", flag_instructions[i].flag);
	} else {
	    emit("    context->flags &= ~%s;\n", flag_instructions[i].flag);
	}
	emit("    context->cycles_left -= 2;\n");
	emit_proctrailer();
    }
}

struct big8_instruction {
    u8 opcode_base;
    char *instruction;
    int is_sta;
    char *block;
} big8_instructions[] = {
    {0x00, "ORA", 0, "    result = context->reg_a |= tmp;\n"},
    {0x20, "AND", 0, "    result = context->reg_a &= tmp;\n"},
    {0x40, "EOR", 0, "    result = context->reg_a ^= tmp;\n"},
    {0x60, "ADC", 0,
     /* FIXME: probably gives bogus N and V flag values in decimal mode */
     "    result = context->reg_a + tmp + (context->flags & FLAG_C);\n"
     "    if (((context->reg_a ^ tmp) & (context->reg_a ^ result)) & 0x80) {\n"
     "        context->flags |= FLAG_V;\n"
     "    } else {\n"
     "        context->flags &= ~FLAG_V;\n"
     "    }\n"
     "    if (context->flags & FLAG_D) {\n"
     "        u8 al;\n"
     "        u8 ah;\n"
     "\n"
     "        al = (context->reg_a & 0x0f) + (tmp & 0x0f) + (context->flags & FLAG_C);\n"
     "        if (al > 9) al += 6;\n"
     "        ah = (context->reg_a >> 4) + (tmp >> 4) + (al > 0x0f);\n"
     "        if (ah > 9) ah += 6;\n"
     "        if (ah > 15) {\n"
     "            context->flags |= FLAG_C;\n"
     "        } else {\n"
     "            context->flags &= ~FLAG_C;\n"
     "        }\n"
     "        context->reg_a = (ah << 4) | (al & 0x0f);\n"
     "    } else {\n"
     "        if (result & 0xff00) {\n"
     "            context->flags |= FLAG_C;\n"
     "        } else {\n"
     "            context->flags &= ~FLAG_C;\n"
     "        }\n"
     "        context->reg_a = result;\n"
     "    }\n"
     "    result &= 0xff;\n"},
    {0x80, "STA", 1, NULL},
    {0xa0, "LDA", 0, "    result = context->reg_a = tmp;\n"},
    {0xc0, "CMP", 0,
     "    result = context->reg_a - tmp;\n"
     "    if (result & 0xff00) {\n"
     "        context->flags &= ~FLAG_C;\n"
     "        result &= 0xff;\n"
     "    } else {\n"
     "        context->flags |= FLAG_C;\n"
     "    }\n"},
    {0xe0, "SBC", 0,
     "    result = (context->reg_a - tmp) - (1 - (context->flags & FLAG_C));\n"
     "    if (result & 0xff00) {\n"
     "        context->flags &= ~FLAG_C;\n"
     "        result &= 0xff;\n"
     "    } else {\n"
     "        context->flags |= FLAG_C;\n"
     "    }\n"
     "    if (((context->reg_a ^ tmp) & (context->reg_a ^ result)) & 0x80) {\n"
     "        context->flags |= FLAG_V;\n"
     "    } else {\n"
     "        context->flags &= ~FLAG_V;\n"
     "    }\n"
     "    if (context->flags & FLAG_D) {\n"
     "        u8 al;\n"
     "        u8 ah;\n"
     "\n"
     "        al = ((context->reg_a & 0x0f) - (tmp & 0x0f)) - (1 - (context->flags & FLAG_C));\n"
     "        if (al & 0x10) al -= 6;\n"
     "        ah = ((context->reg_a >> 4) - (tmp >> 4)) - ((al & 0x10) >> 4);\n"
     "        if (ah & 0x10) ah -= 6;\n"
     "        context->reg_a = (ah << 4) | (al & 0x0f);\n"
     "    } else {\n"
     "        context->reg_a = result;\n"
     "    }\n"},
};

struct amode {
    int cycles;
    char *admode;
    char *preblock;
    char *readblock;
    char *writeblock;
};

struct amode amode_immediate = {
    2, "Imm",
    "\n",
    "    %s = OPFETCH(context);\n",
    "#error Write to Immediate addressing mode\n",
};

struct amode amode_zpage = {
    4, "ZPage",
    "    u8 addr;\n\n    addr = OPFETCH(context);\n",
    "    %s = context->zpage[addr];\n",
    "    context->zpage[addr] = %s;\n",
};

struct amode amode_zpagex = {
    4, "ZPage, X",
    "    u8 addr;\n\n    addr = OPFETCH(context) + context->reg_x;\n",
    "    %s = context->zpage[addr];\n",
    "    context->zpage[addr] = %s;\n"
};

struct amode amode_zpagey = {
    4, "ZPage, Y",
    "    u8 addr;\n\n    addr = OPFETCH(context) + context->reg_y;\n",
    "    %s = context->zpage[addr];\n",
    "    context->zpage[addr] = %s;\n"
};

struct amode amode_abs = {
    5, "Abs",
    "    u16 addr;\n\n    addr = OPFETCH(context);\n    addr += (OPFETCH(context) << 8);\n",
    "    %s = context->readfunc(context->cpu, addr);\n",
    "    context->writefunc(context->cpu, addr, %s);\n"
};

struct amode amode_absx = {
    5, "Abs, X",
    "    u16 addr;\n\n    addr = OPFETCH(context);\n    addr += (OPFETCH(context) << 8);\n    addr += context->reg_x;\n",
    "    %s = context->readfunc(context->cpu, addr);\n",
    "    context->writefunc(context->cpu, addr, %s);\n"
};

struct amode amode_absy = {
    5, "Abs, Y",
    "    u16 addr;\n\n    addr = OPFETCH(context);\n    addr += (OPFETCH(context) << 8);\n    addr += context->reg_y;\n",
    "    %s = context->readfunc(context->cpu, addr);\n",
    "    context->writefunc(context->cpu, addr, %s);\n"
};

struct amode amode_ind = {
    7, "(Ind)",
    "    u8 addr1;\n    u16 addr2;\n\n    addr1 = OPFETCH(context);\n    addr2 = context->zpage[addr1++];\n    addr2 += context->zpage[addr1] << 8;\n",
    "    %s = context->readfunc(context->cpu, addr2);\n",
    "    context->writefunc(context->cpu, addr2, %s);\n"
};

struct amode amode_indx = {
    7, "(Ind, X)",
    "    u8 addr1;\n    u16 addr2;\n\n    addr1 = OPFETCH(context) + context->reg_x;\n    addr2 = context->zpage[addr1++];\n    addr2 += context->zpage[addr1] << 8;\n",
    "    %s = context->readfunc(context->cpu, addr2);\n",
    "    context->writefunc(context->cpu, addr2, %s);\n"
};

struct amode amode_indy = {
    7, "(Ind), Y",
    "    u8 addr1;\n    u16 addr2;\n\n    addr1 = OPFETCH(context);\n    addr2 = context->zpage[addr1++];\n    addr2 += context->zpage[addr1] << 8;\n    addr2 += context->reg_y;\n",
    "    %s = context->readfunc(context->cpu, addr2);\n",
    "    context->writefunc(context->cpu, addr2, %s);\n"
};

struct {
    u8 opcode_delta;
    int is_imm;
    struct amode *amode;
} big8_addmodes[] = {
    {0x09, 1, &amode_immediate},
    {0x05, 0, &amode_zpage},
    {0x15, 0, &amode_zpagex},
    {0x0d, 0, &amode_abs},
    {0x1d, 0, &amode_absx},
    {0x19, 0, &amode_absy},
    {0x12, 0, &amode_ind},
    {0x01, 0, &amode_indx},
    {0x11, 0, &amode_indy},
};

void emit_big8(struct big8_instruction *instr)
{
    int i;
    
    for (i = 0; i < (sizeof(big8_addmodes) / sizeof(big8_addmodes[0])); i++) {
	if (instr->is_sta && big8_addmodes[i].is_imm) {
	    continue;
	}
	
	emit_procheader(instr->opcode_base + big8_addmodes[i].opcode_delta);

	emit("    /* %s %s */\n", instr->instruction, big8_addmodes[i].amode->admode);
	
	if (!instr->is_sta) {
	    emit("    u8 tmp;\n");
	    emit("    u16 result;\n");
	}
	
	emit(big8_addmodes[i].amode->preblock);

	if (!instr->is_sta) {
	    emit(big8_addmodes[i].amode->readblock, "tmp");
	    emit(instr->block);
	    emit("    EVAL_FLAGNZ(context, result);\n");
	} else {
	    emit(big8_addmodes[i].amode->writeblock, "context->reg_a");
	}

	emit("    context->cycles_left -= %d;\n", big8_addmodes[i].amode->cycles);
	
	emit_proctrailer();
    }
}

void emit_big8_instructions(void)
{
    int i;

    for (i = 0; i < (sizeof(big8_instructions) / sizeof(big8_instructions[0])); i++) {
	emit_big8(&big8_instructions[i]);
    }
}

struct {
    u8 opcode;
    char *instruction;
    char *flag;
    char test;
} branch_instructions[] = {
    {0x80, "BRA", NULL, 0},
    {0x10, "BPL", "FLAG_N", '!'},
    {0x30, "BMI", "FLAG_N", ' '},
    {0x50, "BVC", "FLAG_V", '!'},
    {0x70, "BVS", "FLAG_V", ' '},
    {0x90, "BCC", "FLAG_C", '!'},
    {0xb0, "BCS", "FLAG_C", ' '},
    {0xd0, "BNE", "FLAG_Z", '!'},
    {0xf0, "BEQ", "FLAG_Z", ' '},
};

void emit_branch_instructions(void)
{
    int i;

    for (i = 0; i < (sizeof(branch_instructions) / sizeof(branch_instructions[0])); i++) {
	emit_procheader(branch_instructions[i].opcode);
	emit("    /* %s */\n", branch_instructions[i].instruction);
	if (branch_instructions[i].flag) {
	    emit("    if (%c(context->flags & %s)) {\n", branch_instructions[i].test, branch_instructions[i].flag);
	}
	emit("    s8 delta;\n\n    delta = OPFETCH(context);\n");
	emit("    context->pc += delta;\n");
	emit("    context->cycles_left -= 4;\n");
	if (branch_instructions[i].flag) {
	    emit("    } else {\n    context->pc++;\n    context->cycles_left -= 2;\n    }\n");
	}
	emit_proctrailer();
    }
}

void emit_bit_test_branch_instructions(void)
{
    int i;

    for (i = 0; i < 0x10; i++) {
	emit_procheader((i << 4) + 0x0f);
	emit("    /* BB%c%d */\n", 'R' + (i >> 3), i & 7);
	emit("    u8 addr;\n");
	emit("\n");
	emit("    addr = OPFETCH(context);\n");
	emit("\n");
	emit("    if (%c(context->zpage[addr] & 0x%02x)) {\n", (i & 8)? ' ': '!', 1 << (i & 7));
	emit("        s8 delta;\n");
	emit("\n");
	emit("        delta = OPFETCH(context);\n");
	emit("        context->pc += delta;\n");
	emit("        context->cycles_left -= 8;\n");
	emit("    } else {\n");
	emit("        context->pc++;\n");
	emit("        context->cycles_left -= 6;\n");
	emit("    }\n");
	emit_proctrailer();
    }
}

struct {
    u8 opcode;
    char *instruction;
    int is_push;
    char *reg;
    int eval_nz;
} stack_instructions[] = {
    {0x48, "PHA", 1, "context->reg_a", 0},
    {0x08, "PHP", 1, "context->flags", 0},
    {0xda, "PHX", 1, "context->reg_x", 0},
    {0x5a, "PHY", 1, "context->reg_y", 0},
    {0x68, "PLA", 0, "context->reg_a", 1},
    {0x28, "PLP", 0, "context->flags", 0},
    {0xfa, "PLX", 0, "context->reg_x", 1},
    {0x7a, "PLY", 0, "context->reg_y", 1},
};

void emit_stack_instructions(void)
{
    int i;

    for (i = 0; i < (sizeof(stack_instructions) / sizeof(stack_instructions[0])); i++) {
	emit_procheader(stack_instructions[i].opcode);
	emit("    /* %s */\n", stack_instructions[i].instruction);
	if (stack_instructions[i].is_push) {
	    emit("    PUSH_BYTE(context, %s);\n", stack_instructions[i].reg);
	} else {
	    emit("    %s = POP_BYTE(context);\n", stack_instructions[i].reg);
	}
	if (stack_instructions[i].eval_nz) {
	    emit("    EVAL_FLAGNZ(context, %s);\n", stack_instructions[i].reg);
	}
	emit("    context->cycles_left -= 3;\n");
	emit_proctrailer();
    }
}

struct {
    u8 opcode;
    char *instruction;
    char *reg1;
    char *reg2;
    int is_swap;
    int check_flags;
} reg_transfer_instructions[] = {
    {0x22, "SAX", "context->reg_a", "context->reg_x", 1, 0},
    {0x42, "SAY", "context->reg_a", "context->reg_y", 1, 0},
    {0x02, "SXY", "context->reg_x", "context->reg_y", 1, 0},
    {0xaa, "TAX", "context->reg_a", "context->reg_x", 0, 1},
    {0xa8, "TAY", "context->reg_a", "context->reg_y", 0, 1},
    {0xba, "TSX", "context->reg_s", "context->reg_x", 0, 1},
    {0x8a, "TXA", "context->reg_x", "context->reg_a", 0, 1},
    {0x9a, "TXS", "context->reg_x", "context->reg_s", 0, 0},
    {0x98, "TYA", "context->reg_y", "context->reg_a", 0, 1},
};

void emit_reg_transfer_instructions(void)
{
    int i;

    for (i = 0; i < (sizeof(reg_transfer_instructions) / sizeof(reg_transfer_instructions[0])); i++) {
	emit_procheader(reg_transfer_instructions[i].opcode);
	emit("    /* %s */\n", reg_transfer_instructions[i].instruction);
	if (reg_transfer_instructions[i].is_swap) {
	    emit("    u8 tmp;\n");
	    emit("\n");
	    emit("    tmp = %s;\n", reg_transfer_instructions[i].reg1);
	    emit("    %s = %s;\n", reg_transfer_instructions[i].reg1, reg_transfer_instructions[i].reg2);
	    emit("    %s = tmp;\n", reg_transfer_instructions[i].reg2);
	    emit("    context->cycles_left -= 3;\n");
	} else {
	    emit("    %s = %s;\n", reg_transfer_instructions[i].reg2, reg_transfer_instructions[i].reg1);
	    emit("    context->cycles_left -= 2;\n");
	}
	if (reg_transfer_instructions[i].check_flags) {
	    emit("    EVAL_FLAGNZ(context, %s);\n", reg_transfer_instructions[i].reg2);
	}
	emit_proctrailer();
    }
}

struct {
    u8 opcode;
    char *instruction;
    int is_tst;
    struct amode *amode;
    char *rmw_op;
} bit_test_instructions[] = {
    {0x89, "BIT", 0, &amode_immediate, NULL},
    {0x24, "BIT", 0, &amode_zpage, NULL},
    {0x34, "BIT", 0, &amode_zpagex, NULL},
    {0x2c, "BIT", 0, &amode_abs, NULL},
    {0x3c, "BIT", 0, &amode_absx, NULL},
    {0x83, "TST", 1, &amode_zpage, NULL},
    {0xa3, "TST", 1, &amode_zpagex, NULL},
    {0x93, "TST", 1, &amode_abs, NULL},
    {0xb3, "TST", 1, &amode_absx, NULL},
    {0x14, "TRB", 0, &amode_zpage, "&= ~"},
    {0x1c, "TRB", 0, &amode_abs, "&= ~"},
    {0x04, "TSB", 0, &amode_zpage, "|= "},
    {0x0c, "TSB", 0, &amode_abs, "|= "},
};

void emit_bit_test_instructions(void)
{
    int i;
    struct amode *cur_amode;

    for (i = 0; i < (sizeof(bit_test_instructions) / sizeof(bit_test_instructions[0])); i++) {
	cur_amode = bit_test_instructions[i].amode;
    
	emit_procheader(bit_test_instructions[i].opcode);
	emit("    /* %s %s */\n", bit_test_instructions[i].instruction, cur_amode->admode);
	emit("    u8 data;\n");
	if (bit_test_instructions[i].is_tst) {
	    emit("    u8 reg = OPFETCH(context);\n");
	} else {
	    emit("    u8 reg = context->reg_a;\n");
	}
	emit(cur_amode->preblock);
	emit(cur_amode->readblock, "data");
	emit("    context->flags &= ~(0xc0 | FLAG_Z);\n");
	emit("    context->flags |= data & 0xc0;\n");
	emit("    if (!(reg & data)) {\n");
	emit("        context->flags |= FLAG_Z;\n");
	emit("    }\n");
	if (bit_test_instructions[i].rmw_op) {
	    emit("    data %sreg;\n", bit_test_instructions[i].rmw_op);
	    emit(cur_amode->writeblock, "data");
	    emit("    context->cycles_left -= %d;\n", cur_amode->cycles + 2);
	} else if (bit_test_instructions[i].is_tst) {
	    emit("    context->cycles_left -= %d;\n", cur_amode->cycles + 3);
	} else {
	    emit("    context->cycles_left -= %d;\n", cur_amode->cycles);
	}
	emit_proctrailer();
    }
}

struct amode amode_rega = {
    0, "Acc",
    "\n",
    "    %s = context->reg_a;\n",
    "    context->reg_a = %s;\n",
};

struct amode amode_regx = {
    0, "",
    "\n",
    "    %s = context->reg_x;\n",
    "    context->reg_x = %s;\n",
};

struct amode amode_regy = {
    0, "",
    "\n",
    "    %s = context->reg_y;\n",
    "    context->reg_y = %s;\n",
};

struct {
    u8 opcode;
    char *instruction;
    char operation;
    struct amode *amode;
} inc_dec_instructions[] = {
    {0xe6, "INC", '+', &amode_zpage},
    {0xf6, "INC", '+', &amode_zpagex},
    {0xee, "INC", '+', &amode_abs},
    {0xfe, "INC", '+', &amode_absx},
    {0x1a, "INC", '+', &amode_rega},
    {0xe8, "INX", '+', &amode_regx},
    {0xc8, "INY", '+', &amode_regy},
    {0xc6, "DEC", '-', &amode_zpage},
    {0xd6, "DEC", '-', &amode_zpagex},
    {0xce, "DEC", '-', &amode_abs},
    {0xde, "DEC", '-', &amode_absx},
    {0x3a, "DEC", '-', &amode_rega},
    {0xca, "DEX", '-', &amode_regx},
    {0x88, "DEY", '-', &amode_regy},
};

void emit_inc_dec_instructions(void)
{
    int i;
    struct amode *cur_amode;

    for (i = 0; i < (sizeof(inc_dec_instructions) / sizeof(inc_dec_instructions[0])); i++) {
	cur_amode = inc_dec_instructions[i].amode;
    
	emit_procheader(inc_dec_instructions[i].opcode);
	emit("    /* %s %s */\n", inc_dec_instructions[i].instruction, cur_amode->admode);
	emit("    u8 data;\n");
	emit(cur_amode->preblock);
	emit(cur_amode->readblock, "data");
	emit("    data %c= 1;\n", inc_dec_instructions[i].operation);
	emit(cur_amode->writeblock, "data");
	emit("    EVAL_FLAGNZ(context, data);\n");
	emit("    context->cycles_left -= %d;\n", cur_amode->cycles + 2);
	emit_proctrailer();
    }
}

struct {
    u8 opcode;
    char *instruction;
    int is_store;
    int cycle_delta;
    char *reg;
    struct amode *amode;
} load_store_instructions[] = {
    {0xa2, "LDX", 0, 0, "context->reg_x", &amode_immediate},
    {0xa6, "LDX", 0, 0, "context->reg_x", &amode_zpage},
    {0xb6, "LDX", 0, 0, "context->reg_x", &amode_zpagey},
    {0xae, "LDX", 0, 0, "context->reg_x", &amode_abs},
    {0xbe, "LDX", 0, 0, "context->reg_x", &amode_absy},
    {0xa0, "LDY", 0, 0, "context->reg_y", &amode_immediate},
    {0xa4, "LDY", 0, 0, "context->reg_y", &amode_zpage},
    {0xb4, "LDY", 0, 0, "context->reg_y", &amode_zpagex},
    {0xac, "LDY", 0, 0, "context->reg_y", &amode_abs},
    {0xbc, "LDY", 0, 0, "context->reg_y", &amode_absx},
    {0x62, "CLA", 1, 2, "0", &amode_rega},
    {0x82, "CLX", 1, 2, "0", &amode_regx},
    {0xc2, "CLY", 1, 2, "0", &amode_regy},
    {0x86, "STX", 1, 0, "context->reg_x", &amode_zpage},
    {0x96, "STX", 1, 0, "context->reg_x", &amode_zpagey},
    {0x8e, "STX", 1, 0, "context->reg_x", &amode_abs},
    {0x84, "STY", 1, 0, "context->reg_y", &amode_zpage},
    {0x94, "STY", 1, 0, "context->reg_y", &amode_zpagex},
    {0x8c, "STY", 1, 0, "context->reg_y", &amode_abs},
    {0x64, "STZ", 1, 0, "0", &amode_zpage},
    {0x74, "STZ", 1, 0, "0", &amode_zpagex},
    {0x9c, "STZ", 1, 0, "0", &amode_abs},
    {0x9e, "STZ", 1, 0, "0", &amode_absx},
};

void emit_load_store_instructions(void)
{
    int i;
    struct amode *cur_amode;

    for (i = 0; i < (sizeof(load_store_instructions) / sizeof(load_store_instructions[0])); i++) {
	cur_amode = load_store_instructions[i].amode;
    
	emit_procheader(load_store_instructions[i].opcode);
	emit("    /* %s %s */\n", load_store_instructions[i].instruction, cur_amode->admode);
	emit(cur_amode->preblock);
	if (!(load_store_instructions[i].is_store)) {
	    emit(cur_amode->readblock, load_store_instructions[i].reg);
	    emit("    EVAL_FLAGNZ(context, %s);\n", load_store_instructions[i].reg);
	} else {
	    emit(cur_amode->writeblock, load_store_instructions[i].reg);
	}
	emit("    context->cycles_left -= %d;\n", cur_amode->cycles + load_store_instructions[i].cycle_delta);
	emit_proctrailer();
    }
}

struct shift_rotate_instruction {
    u8 opcode;
    char *instruction;
    int is_rotate;
    int is_left;
} shift_rotate_instructions[] = {
    {0x00, "ASL", 0, 1},
    {0x20, "ROL", 1, 1},
    {0x40, "LSR", 0, 0},
    {0x60, "ROR", 1, 0},
    {0x00, NULL, 0, 0}, /* must be last entry */
};

struct shift_rotate_amode {
    u8 opcode_delta;
    struct amode *amode;
} shift_rotate_amodes[] = {
    {0x0a, &amode_rega},
    {0x06, &amode_zpage},
    {0x16, &amode_zpagex},
    {0x0e, &amode_abs},
    {0x1e, &amode_absx},
    {0x00, NULL}, /* must be last entry */
};

void emit_shift_rotate_instructions(void)
{
    struct shift_rotate_instruction *cur_instr;
    struct shift_rotate_amode *cur_amode;
    
    for (cur_instr = shift_rotate_instructions; cur_instr->instruction; cur_instr++) {
	for (cur_amode = shift_rotate_amodes; cur_amode->amode; cur_amode++) {
	    emit_procheader(cur_instr->opcode + cur_amode->opcode_delta);
	    emit("    /* %s %s */\n", cur_instr->instruction, cur_amode->amode->admode);
	    emit("    u8 src_data;\n");
	    emit("    u8 dst_data;\n");
	    emit(cur_amode->amode->preblock);
	    emit(cur_amode->amode->readblock, "src_data");
	    emit("    dst_data = src_data %s 1;\n", cur_instr->is_left? "<<": ">>");
	    if (cur_instr->is_rotate) {
		if (cur_instr->is_left) {
		    emit("    dst_data |= context->flags & FLAG_C;\n");
		} else {
		    emit("    dst_data |= context->flags << 7;\n");
		}
	    }
	    emit(cur_amode->amode->writeblock, "dst_data");
	    emit("    context->flags &= ~FLAG_C;\n");
	    if (cur_instr->is_left) {
		emit("    context->flags |= src_data >> 7;\n");
	    } else {
		emit("    context->flags |= src_data & FLAG_C;\n");
	    }
	    emit("    EVAL_FLAGNZ(context, dst_data);\n");
	    emit("    context->cycles_left -= %d;\n", cur_amode->amode->cycles + 2);
	    emit_proctrailer();
	}
    }
}

void emit_bit_manipulation_instructions(void)
{
    int i;

    for (i = 0; i < 16; i++) {
	emit_procheader((i << 4) + 0x07);
	emit("    /* %cMB%d */\n", 'R' + (i >> 3), i & 7);
	emit("    u8 addr;\n");
	emit("\n");
	emit("    addr = OPFETCH(context);\n");
	emit("\n");
	emit("    context->zpage[addr] %s%d;\n", (i & 8)? "|= ": "&= ~", 1 << (i & 7));
	emit("\n");
	emit("    context->cycles_left -= 7;\n");
	emit_proctrailer();
    }
}

struct compare_amode {
    u8 opcode_delta;
    struct amode *amode;
} compare_amodes[] = {
    {0xc0, &amode_immediate},
    {0xc4, &amode_zpage},
    {0xcc, &amode_abs},
    {0x00, NULL},
};

void emit_compare_instructions(void)
{
    int i;
    struct compare_amode *cur_amode;

    for (i = 0; i < 2; i++) {
	for (cur_amode = compare_amodes; cur_amode->amode; cur_amode++) {
	    emit_procheader((i << 5) + cur_amode->opcode_delta);
	    emit("    /* CP%c %s */\n", 'Y' - i, cur_amode->amode->admode);
	    emit("    u8 tmp;\n");
	    emit("    u8 result;\n");
	    
	    emit(cur_amode->amode->preblock);
	    emit(cur_amode->amode->readblock, "tmp");
	    
	    emit("    result = context->reg_%c - tmp;\n", 'y' - i);
	    emit("    if (context->reg_%c < tmp) {\n", 'y' - i);
	    emit("        context->flags &= ~FLAG_C;\n");
	    emit("    } else {\n");
	    emit("        context->flags |= FLAG_C;\n");
	    emit("    }\n");
	    emit("    EVAL_FLAGNZ(context, result);\n");
	    emit("    context->cycles_left -= %d;\n", cur_amode->amode->cycles);
	    emit_proctrailer();
	}
    }
}

struct block_instruction {
    u8 opcode;
    char *instruction;
    char *src_op1;
    char *src_op2;
    char *dst_op1;
    char *dst_op2;
} block_instructions[] = {
    {0xf3, "TAI", "++", "--", "++", "++"},
    {0xc3, "TDD", "--", "--", "--", "--"},
    {0xe3, "TIA", "++", "++", "++", "--"},
    {0x73, "TII", "++", "++", "++", "++"},
    {0xd3, "TIN", "++", "++", "", ""},
    {0x00, NULL, NULL, NULL, NULL, NULL}, /* must be last entry */
};

void emit_block_instructions(void)
{
    struct block_instruction *cur_instr;

    for (cur_instr = block_instructions; cur_instr->instruction; cur_instr++) {
	emit_procheader(cur_instr->opcode);
	emit("    /* %s */\n", cur_instr->instruction);
	emit("    u16 source;\n");
	emit("    u16 dest;\n");
	emit("    u32 length;\n");
	emit("\n");
	emit("    source = OPFETCH(context);\n");
	emit("    source |= OPFETCH(context) << 8;\n");
	emit("    dest = OPFETCH(context);\n");
	emit("    dest |= OPFETCH(context) << 8;\n");
	emit("    length = OPFETCH(context);\n");
	emit("    length |= OPFETCH(context) << 8;\n");
	emit("\n");
	emit("    if (!length) length = 0x10000;\n");
	emit("    context->cycles_left -= (17 + (length * 6));\n");
	emit("\n");
	emit("    for (;;) {\n");
	emit("        context->writefunc(context->cpu, dest%s, context->readfunc(context->cpu, source%s));\n", cur_instr->dst_op1, cur_instr->src_op1);
	emit("        if (!--length) break;\n");
	emit("        context->writefunc(context->cpu, dest%s, context->readfunc(context->cpu, source%s));\n", cur_instr->dst_op2, cur_instr->src_op2);
	emit("        if (!--length) break;\n");
	emit("    }\n");
	emit_proctrailer();
    }
}

struct misc_instruction {
    u8 opcode;
    char *instruction;
} misc_instructions[] = {
    /* call/return group */
    {0x4c, "    /* JMP Abs */\n"
     "    u16 addr;\n"
     "\n"
     "    addr = OPFETCH(context);\n"
     "    addr |= OPFETCH(context) << 8;\n"
     "    context->pc = addr;\n"
     "    context->cycles_left -= 4;\n"},
    {0x6c, "    /* JMP Ind */\n"
     "    u16 addr;\n"
     "\n"
     "    addr = OPFETCH(context);\n"
     "    addr |= OPFETCH(context) << 8;\n"
     "    context->pc = context->readfunc(context->cpu, addr++);\n"
     "    context->pc |= context->readfunc(context->cpu, addr) << 8;\n"
     "    context->cycles_left -= 7;\n"},
    {0x7c, "    /* JMP Ind, X */\n"
     "    u16 addr;\n"
     "\n"
     "    addr = OPFETCH(context);\n"
     "    addr |= OPFETCH(context) << 8;\n"
     "    addr += context->reg_x;\n"
     "    context->pc = context->readfunc(context->cpu, addr++);\n"
     "    context->pc |= context->readfunc(context->cpu, addr) << 8;\n"
     "    context->cycles_left -= 7;\n"},
    {0x20, "    /* JSR */\n"
     "    u16 addr;\n"
     "\n"
     "    addr = OPFETCH(context);\n"
     "    PUSH_BYTE(context, context->pc >> 8);\n"
     "    PUSH_BYTE(context, context->pc & 0xff);\n"
     "    addr |= OPFETCH(context) << 8;\n"
     "    context->pc = addr;\n"
     "    context->cycles_left -= 7;\n"},
    {0x44, "    /* BSR */\n"
     "    s8 delta;\n"
     "\n"
     "    PUSH_BYTE(context, context->pc >> 8);\n"
     "    PUSH_BYTE(context, context->pc & 0xff);\n"
     "    delta = OPFETCH(context);\n"
     "    context->pc += delta;\n"
     "    context->cycles_left -= 8;\n"},
    {0x60, "    /* RTS */\n"
     "\n"
     "    context->pc = POP_BYTE(context);\n"
     "    context->pc |= POP_BYTE(context) << 8;\n"
     "    context->pc++;\n"
     "    context->cycles_left -= 7;\n"},
    {0x40, "    /* RTI */\n"
     "\n"
     "    context->flags = POP_BYTE(context);\n"
     "    context->pc = POP_BYTE(context);\n"
     "    context->pc |= POP_BYTE(context) << 8;\n"
     "    context->cycles_left -= 7;\n"},
    {0x00, "    /* BRK */\n"
     "\n"
     "    context->pc++;\n"
     "    PUSH_BYTE(context, context->pc >> 8);\n"
     "    PUSH_BYTE(context, context->pc & 0xff);\n"
     "    PUSH_BYTE(context, context->flags | FLAG_B);\n"
     "    context->flags |= FLAG_I;\n"
     "    context->flags &= ~FLAG_D;\n"
     "    context->pc = context->readfunc(context->cpu, VECTOR_BRK_LO);\n"
     "    context->pc |= context->readfunc(context->cpu, VECTOR_BRK_HI) << 8;\n"
     "    context->cycles_left -= 8;\n"},
    {0xea, "    /* NOP */\n"},
    {0xd4, "    /* CSH */\n"},
    {0x54, "    /* CSL */\n"},
    {0x03, "    /* ST0 */\n"
     "    void pce_st0_hack(u8);\n"
     "    pce_st0_hack(OPFETCH(context));\n"
     "    context->cycles_left -= 4;\n"},
    {0x13, "    /* ST1 */\n"
     "    void pce_st1_hack(u8);\n"
     "    pce_st1_hack(OPFETCH(context));\n"
     "    context->cycles_left -= 4;\n"},
    {0x23, "    /* ST2 */\n"
     "    void pce_st2_hack(u8);\n"
     "    pce_st2_hack(OPFETCH(context));\n"
     "    context->cycles_left -= 4;\n"},
    {0x43, "    /* TMAi */\n"
     "    u8 i;\n"
     "\n"
     "    i = OPFETCH(context);\n"
     "    if (i & 0x80) { context->reg_a = context->mpr[7]; }\n"
     "    if (i & 0x40) { context->reg_a = context->mpr[6]; }\n"
     "    if (i & 0x20) { context->reg_a = context->mpr[5]; }\n"
     "    if (i & 0x10) { context->reg_a = context->mpr[4]; }\n"
     "    if (i & 0x08) { context->reg_a = context->mpr[3]; }\n"
     "    if (i & 0x04) { context->reg_a = context->mpr[2]; }\n"
     "    if (i & 0x02) { context->reg_a = context->mpr[1]; }\n"
     "    if (i & 0x01) { context->reg_a = context->mpr[0]; }\n"
     "    context->cycles_left -= 4;\n"},
    {0x53, "    /* TAMi */\n"
     "    u8 i;\n"
     "\n"
     "    i = OPFETCH(context);\n"
     "    if (i & 0x80) { context->mpr[7] = context->reg_a; }\n"
     "    if (i & 0x40) { context->mpr[6] = context->reg_a; }\n"
     "    if (i & 0x20) { context->mpr[5] = context->reg_a; }\n"
     "    if (i & 0x10) { context->mpr[4] = context->reg_a; }\n"
     "    if (i & 0x08) { context->mpr[3] = context->reg_a; }\n"
     "    if (i & 0x04) { context->mpr[2] = context->reg_a; }\n"
     "    if (i & 0x02) { context->mpr[1] = context->reg_a; }\n"
     "    if (i & 0x01) { context->mpr[0] = context->reg_a; }\n"
     "    context->cycles_left -= 5;\n"},
    {0xf4, "    /* SET */\n"
     /* NOTE: this uses some black magic to determine what to do */
     "    u8 i;\n"
     "    u8 tmp;\n"
     "\n"
     "    i = OPFETCH(context);\n"
     /* the if is checking for ADC, AND, EOR, and ORA instructions */
     "    if (((i & 0x83) == 0x01) || ((i & 0x9f) == 0x12)) {\n"
     "        tmp = context->reg_a;\n"
     "        context->reg_a = context->zpage[context->reg_x];\n"
     "        cpu6280_decode[i](context);\n"
     "        context->zpage[context->reg_x] = context->reg_a;\n"
     "        context->reg_a = tmp;\n"
     "        context->cycles_left -= 5;\n"
     "    } else {\n"
     "        cpu6280_decode[i](context);\n"
     "        context->cycles_left -= 2;\n"
     "    }\n"},
    {0x00, NULL}, /* must be last entry */
};

void emit_misc_instructions(void)
{
    struct misc_instruction *cur_instr;

    for (cur_instr = misc_instructions; cur_instr->instruction; cur_instr++) {
	emit_procheader(cur_instr->opcode);
	emit(cur_instr->instruction);
	emit_proctrailer();
    }
}

instr_emitter_t emitter_list[] = {
    emit_flag_instructions,
    emit_big8_instructions,
    emit_branch_instructions,
    emit_bit_test_branch_instructions,
    emit_stack_instructions,
    emit_reg_transfer_instructions,
    emit_bit_test_instructions,
    emit_inc_dec_instructions,
    emit_load_store_instructions,
    emit_shift_rotate_instructions,
    emit_bit_manipulation_instructions,
    emit_compare_instructions,
    emit_block_instructions,
    emit_misc_instructions,
    NULL, /* must be last entry */
};

void build_core(void)
{
    int i;

    for (i = 0; emitter_list[i]; i++) {
	emitter_list[i]();
    }
}

int main(int argc, char *argv[])
{
    if (argc != 2) {
	printf("usage: %s <outfile>\n", argv[0]);
	return 1;
    }

    outfile = fopen(argv[1], "w");
    if (!outfile) {
	perror("Unable to open file for writing");
	return 1;
    }

    emit_file_header(argv[1]);

    clear_handlers();

    build_core();
    
    emit_decode_table();
    
    emit_file_footer();

    if (fclose(outfile) == EOF) {
	perror("Closing file");
	return 1;
    }
    
    return 0;
}

/*
 * $Log: build_cpu6280.c,v $
 * Revision 1.29  2000/02/11 03:26:09  nyef
 * added SET instruction
 *
 * Revision 1.28  2000/01/27 01:27:04  nyef
 * added some flag manipulation whatnot to the BRK instruction
 *
 * Revision 1.27  2000/01/27 00:15:50  nyef
 * changed emit_procheader() to assign the current handler to an opcode
 * changed emit_procheader() to output the opcode as a comment
 *
 * Revision 1.26  2000/01/24 05:01:46  nyef
 * added (not strictly tested) implementation of ADC in decimal mode
 *
 * Revision 1.25  2000/01/24 04:27:55  nyef
 * added (untested) implementation of SBC in decimal mode
 *
 * Revision 1.24  2000/01/24 03:43:43  nyef
 * fixed a bug with the bit test instructions
 *
 * Revision 1.23  2000/01/24 03:43:27  nyef
 * added (untested) preliminary support for ADC and SBC
 *
 * Revision 1.22  2000/01/24 00:45:33  nyef
 * added (untested) NOP, CSH, CSL, ST0, ST1, ST2, TAMi, and TMAi
 *
 * Revision 1.21  2000/01/24 00:13:14  nyef
 * added (untested) block transfer instructions
 *
 * Revision 1.20  2000/01/23 23:03:51  nyef
 * added JMP, JSR, BSR, RTS, RTI, and BRK instructions
 *
 * Revision 1.19  2000/01/23 20:00:04  nyef
 * added CPX and CPY instructions
 *
 * Revision 1.18  2000/01/23 18:33:15  nyef
 * fixed problem with debug comments on RMBi, SMBi, BBRi, and BBSi instructions
 *
 * Revision 1.17  2000/01/23 18:30:26  nyef
 * added RMBi and SMBi instructions
 *
 * Revision 1.16  2000/01/23 16:49:14  nyef
 * added shift/rotate instructions
 *
 * Revision 1.15  2000/01/23 05:50:51  nyef
 * added LDX, LDY, STX, STY, CLA, CLX, CLY, and STZ instructions
 *
 * Revision 1.14  2000/01/23 01:10:49  nyef
 * added increment and decrement instructions
 *
 * Revision 1.13  2000/01/22 22:58:41  nyef
 * added bit test instructions (BIT, TST, TSB, TRB)
 *
 * Revision 1.12  2000/01/22 19:19:13  nyef
 * separated amodes out into separate structures from the "big 8" list
 *
 * Revision 1.11  2000/01/22 16:23:47  nyef
 * added register transfer instructions
 *
 * Revision 1.10  2000/01/22 05:10:09  nyef
 * added PHA, PHP, PHX, PHY, PLA, PLP, PLX, and PLY instructions
 *
 * Revision 1.9  2000/01/22 04:09:54  nyef
 * added BBRi and BBSi instructions
 *
 * Revision 1.8  2000/01/22 03:33:36  nyef
 * added branch instructions (except BBRi and BBSi)
 *
 * Revision 1.7  2000/01/22 02:25:39  nyef
 * added (Indirect); (Indirect, X); and (Indirect), Y "big 8" addressing modes
 *
 * Revision 1.6  2000/01/21 02:27:38  nyef
 * added Absolute; Absolute, X; and Absolute, Y "big 8" addressing modes
 *
 * Revision 1.5  2000/01/20 04:07:56  nyef
 * added Immediate and ZPage, X "big 8" addressing modes
 *
 * Revision 1.4  2000/01/19 04:11:01  nyef
 * added CMP "big 8" instruction
 *
 * Revision 1.3  2000/01/19 01:34:57  nyef
 * added all "big 8" instructions other than ADC, SBC, and CMP (zpage only)
 *
 * Revision 1.2  2000/01/17 03:03:42  nyef
 * added code to create opcode handlers
 * added all flag instructions other than SET
 *
 * Revision 1.1  2000/01/17 01:04:33  nyef
 * Initial revision
 *
 */
