/*
 * build_cpu6502.c
 *
 * build the cpu6502 core opcode handlers
 */

/* $Id: build_cpu6502.c,v 1.1 2000/05/01 00:34:17 nyef Exp $ */

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>

#include "types.h"

FILE *outfile;

int handlers[0x100];
int cur_handler;

const char file_header[] = "/*\n * %s\n *\n * opcode handlers for cpu6502 core\n *\n * MACHINE GENERATED. DO NOT EDIT.\n */\n\n#include \"cpu6502int.h\"\n\n";

const char file_footer[] = "/* EOF */\n";

typedef void (*instr_emitter_t)(void);

void emit(const char *format, ...)
{
    va_list args;

    va_start(args, format);
    vfprintf(outfile, format, args);
    va_end(args);
}

void emit_file_header(char *filename)
{
    emit(file_header, filename);
}

void emit_file_footer(void)
{
    emit(file_footer);
}

void clear_handlers(void)
{
    int i;

    for (i = 0; i < 0x100; i++) {
	handlers[i] = 0;
    }

    cur_handler = 1;
}

void emit_decode_table(void)
{
    int i;

    emit("ophandler_t cpu6502_decode[0x100] = {\n");
    
    for (i = 0; i < 0x100; i++) {
	if ((i & 3) == 0) {
	    emit("   ");
	}
	emit(" cpu6502_op_%d,", handlers[i]);
	if ((i & 3) == 3) {
	    emit(" /* 0x%02x - 0x%02x */\n", i & 0xfc, i);
	}
    }

    emit("};\n\n");
}

void emit_procheader(u8 opcode)
{
    emit("void cpu6502_op_%d(struct cpu6502_context *context)\n", cur_handler);
    emit("{\n");
    emit("    /* 0x%02x */\n", opcode);
    handlers[opcode] = cur_handler;
}

void emit_proctrailer(void)
{
    emit("}\n\n");

    cur_handler++;
}

struct {
    u8 opcode;
    char *instruction;
    char *flag;
    int set;
} flag_instructions[] = {
    {0x18, "CLC", "FLAG_C", 0},
    {0x38, "SEC", "FLAG_C", 1},
    {0xd8, "CLD", "FLAG_D", 0},
    {0xf8, "SED", "FLAG_D", 1},
    {0x58, "CLI", "FLAG_I", 0},
    {0x78, "SEI", "FLAG_I", 1},
    {0xb8, "CLV", "FLAG_V", 0},
};

void emit_flag_instructions(void)
{
    int i;

    for (i = 0; i < (sizeof(flag_instructions) / sizeof(flag_instructions[0])); i++) {
	emit_procheader(flag_instructions[i].opcode);
	emit("    /* %s */\n", flag_instructions[i].instruction);
	if (flag_instructions[i].set) {
	    emit("    context->flags |= %s;\n", flag_instructions[i].flag);
	} else {
	    emit("    context->flags &= ~%s;\n", flag_instructions[i].flag);
	}
	emit("    context->cycles_left -= 2;\n");
	emit_proctrailer();
    }
}

struct big8_instruction {
    u8 opcode_base;
    char *instruction;
    int is_sta;
    char *block;
} big8_instructions[] = {
    {0x00, "ORA", 0, "    result = context->reg_a |= tmp;\n"},
    {0x20, "AND", 0, "    result = context->reg_a &= tmp;\n"},
    {0x40, "EOR", 0, "    result = context->reg_a ^= tmp;\n"},
    {0x60, "ADC", 0,
     /* FIXME: probably gives bogus N and V flag values in decimal mode */
     "    result = context->reg_a + tmp + (context->flags & FLAG_C);\n"
     "    if (((result ^ tmp) & (context->reg_a ^ result)) & 0x80) {\n"
     "        context->flags |= FLAG_V;\n"
     "    } else {\n"
     "        context->flags &= ~FLAG_V;\n"
     "    }\n"
     "    if (context->flags & FLAG_D) {\n"
     "        u8 al;\n"
     "        u8 ah;\n"
     "\n"
     "        al = (context->reg_a & 0x0f) + (tmp & 0x0f) + (context->flags & FLAG_C);\n"
     "        if (al > 9) al += 6;\n"
     "        ah = (context->reg_a >> 4) + (tmp >> 4) + (al > 0x0f);\n"
     "        if (ah > 9) ah += 6;\n"
     "        if (ah > 15) {\n"
     "            context->flags |= FLAG_C;\n"
     "        } else {\n"
     "            context->flags &= ~FLAG_C;\n"
     "        }\n"
     "        context->reg_a = (ah << 4) | (al & 0x0f);\n"
     "    } else {\n"
     "        if (result & 0xff00) {\n"
     "            context->flags |= FLAG_C;\n"
     "        } else {\n"
     "            context->flags &= ~FLAG_C;\n"
     "        }\n"
     "        context->reg_a = result;\n"
     "    }\n"
     "    result &= 0xff;\n"},
    {0x80, "STA", 1, NULL},
    {0xa0, "LDA", 0, "    result = context->reg_a = tmp;\n"},
    {0xc0, "CMP", 0,
     "    result = context->reg_a - tmp;\n"
     "    if (result & 0xff00) {\n"
     "        context->flags &= ~FLAG_C;\n"
     "        result &= 0xff;\n"
     "    } else {\n"
     "        context->flags |= FLAG_C;\n"
     "    }\n"},
    {0xe0, "SBC", 0,
     "    result = (context->reg_a - tmp) - (1 - (context->flags & FLAG_C));\n"
     "    if (result & 0xff00) {\n"
     "        context->flags &= ~FLAG_C;\n"
     "        result &= 0xff;\n"
     "    } else {\n"
     "        context->flags |= FLAG_C;\n"
     "    }\n"
     "    if (((context->reg_a ^ tmp) & (tmp ^ result)) & 0x80) {\n"
     "        context->flags |= FLAG_V;\n"
     "    } else {\n"
     "        context->flags &= ~FLAG_V;\n"
     "    }\n"
     "    if (context->flags & FLAG_D) {\n"
     "        u8 al;\n"
     "        u8 ah;\n"
     "\n"
     "        al = ((context->reg_a & 0x0f) - (tmp & 0x0f)) - (1 - (context->flags & FLAG_C));\n"
     "        if (al & 0x10) al -= 6;\n"
     "        ah = ((context->reg_a >> 4) - (tmp >> 4)) - ((al & 0x10) >> 4);\n"
     "        if (ah & 0x10) ah -= 6;\n"
     "        context->reg_a = (ah << 4) | (al & 0x0f);\n"
     "    } else {\n"
     "        context->reg_a = result;\n"
     "    }\n"},
};

struct amode {
    int cycles;
    char *admode;
    char *preblock;
    char *readblock;
    char *writeblock;
};

struct amode amode_immediate = {
    2, "Imm",
    "\n",
    "    %s = OPFETCH(context);\n",
    "#error Write to Immediate addressing mode\n",
};

struct amode amode_zpage = {
    4, "ZPage",
    "    u8 addr;\n\n    addr = OPFETCH(context);\n",
    "    %s = context->zpage[addr];\n",
    "    context->zpage[addr] = %s;\n",
};

struct amode amode_zpagex = {
    4, "ZPage, X",
    "    u8 addr;\n\n    addr = OPFETCH(context) + context->reg_x;\n",
    "    %s = context->zpage[addr];\n",
    "    context->zpage[addr] = %s;\n"
};

struct amode amode_zpagey = {
    4, "ZPage, Y",
    "    u8 addr;\n\n    addr = OPFETCH(context) + context->reg_y;\n",
    "    %s = context->zpage[addr];\n",
    "    context->zpage[addr] = %s;\n"
};

struct amode amode_abs = {
    5, "Abs",
    "    u16 addr;\n\n    addr = OPFETCH(context);\n    addr += (OPFETCH(context) << 8);\n",
    "    %s = context->readfunc(context->cpu, addr);\n",
    "    context->writefunc(context->cpu, addr, %s);\n"
};

struct amode amode_absx = {
    5, "Abs, X",
    "    u16 addr;\n\n    addr = OPFETCH(context);\n    addr += (OPFETCH(context) << 8);\n    addr += context->reg_x;\n",
    "    %s = context->readfunc(context->cpu, addr);\n",
    "    context->writefunc(context->cpu, addr, %s);\n"
};

struct amode amode_absy = {
    5, "Abs, Y",
    "    u16 addr;\n\n    addr = OPFETCH(context);\n    addr += (OPFETCH(context) << 8);\n    addr += context->reg_y;\n",
    "    %s = context->readfunc(context->cpu, addr);\n",
    "    context->writefunc(context->cpu, addr, %s);\n"
};

struct amode amode_ind = {
    7, "(Ind)",
    "    u8 addr1;\n    u16 addr2;\n\n    addr1 = OPFETCH(context);\n    addr2 = context->zpage[addr1++];\n    addr2 += context->zpage[addr1] << 8;\n",
    "    %s = context->readfunc(context->cpu, addr2);\n",
    "    context->writefunc(context->cpu, addr2, %s);\n"
};

struct amode amode_indx = {
    7, "(Ind, X)",
    "    u8 addr1;\n    u16 addr2;\n\n    addr1 = OPFETCH(context) + context->reg_x;\n    addr2 = context->zpage[addr1++];\n    addr2 += context->zpage[addr1] << 8;\n",
    "    %s = context->readfunc(context->cpu, addr2);\n",
    "    context->writefunc(context->cpu, addr2, %s);\n"
};

struct amode amode_indy = {
    7, "(Ind), Y",
    "    u8 addr1;\n    u16 addr2;\n\n    addr1 = OPFETCH(context);\n    addr2 = context->zpage[addr1++];\n    addr2 += context->zpage[addr1] << 8;\n    addr2 += context->reg_y;\n",
    "    %s = context->readfunc(context->cpu, addr2);\n",
    "    context->writefunc(context->cpu, addr2, %s);\n"
};

struct {
    u8 opcode_delta;
    int is_imm;
    struct amode *amode;
} big8_addmodes[] = {
    {0x09, 1, &amode_immediate},
    {0x05, 0, &amode_zpage},
    {0x15, 0, &amode_zpagex},
    {0x0d, 0, &amode_abs},
    {0x1d, 0, &amode_absx},
    {0x19, 0, &amode_absy},
    {0x01, 0, &amode_indx},
    {0x11, 0, &amode_indy},
};

void emit_big8(struct big8_instruction *instr)
{
    int i;
    
    for (i = 0; i < (sizeof(big8_addmodes) / sizeof(big8_addmodes[0])); i++) {
	if (instr->is_sta && big8_addmodes[i].is_imm) {
	    continue;
	}
	
	emit_procheader(instr->opcode_base + big8_addmodes[i].opcode_delta);

	emit("    /* %s %s */\n", instr->instruction, big8_addmodes[i].amode->admode);
	
	if (!instr->is_sta) {
	    emit("    u8 tmp;\n");
	    emit("    u16 result;\n");
	}
	
	emit(big8_addmodes[i].amode->preblock);

	if (!instr->is_sta) {
	    emit(big8_addmodes[i].amode->readblock, "tmp");
	    emit(instr->block);
	    emit("    EVAL_FLAGNZ(context, result);\n");
	} else {
	    emit(big8_addmodes[i].amode->writeblock, "context->reg_a");
	}

	emit("    context->cycles_left -= %d;\n", big8_addmodes[i].amode->cycles);
	
	emit_proctrailer();
    }
}

void emit_big8_instructions(void)
{
    int i;

    for (i = 0; i < (sizeof(big8_instructions) / sizeof(big8_instructions[0])); i++) {
	emit_big8(&big8_instructions[i]);
    }
}

struct {
    u8 opcode;
    char *instruction;
    char *flag;
    char test;
} branch_instructions[] = {
    {0x10, "BPL", "FLAG_N", '!'},
    {0x30, "BMI", "FLAG_N", ' '},
    {0x50, "BVC", "FLAG_V", '!'},
    {0x70, "BVS", "FLAG_V", ' '},
    {0x90, "BCC", "FLAG_C", '!'},
    {0xb0, "BCS", "FLAG_C", ' '},
    {0xd0, "BNE", "FLAG_Z", '!'},
    {0xf0, "BEQ", "FLAG_Z", ' '},
};

void emit_branch_instructions(void)
{
    int i;

    for (i = 0; i < (sizeof(branch_instructions) / sizeof(branch_instructions[0])); i++) {
	emit_procheader(branch_instructions[i].opcode);
	emit("    /* %s */\n", branch_instructions[i].instruction);
	emit("    if (%c(context->flags & %s)) {\n", branch_instructions[i].test, branch_instructions[i].flag);
	emit("        s8 delta;\n\n    delta = OPFETCH(context);\n");
	emit("        context->pc += delta;\n");
	emit("        context->cycles_left -= 3;\n");
	emit("    } else {\n");
	emit("        context->pc++;\n");
	emit("        context->cycles_left -= 2;\n");
	emit("    }\n");
	emit_proctrailer();
    }
}

struct {
    u8 opcode;
    char *instruction;
    int is_push;
    char *reg;
    int eval_nz;
} stack_instructions[] = {
    {0x48, "PHA", 1, "context->reg_a", 0},
    {0x08, "PHP", 1, "context->flags", 0},
    {0x68, "PLA", 0, "context->reg_a", 1},
    {0x28, "PLP", 0, "context->flags", 0},
};

void emit_stack_instructions(void)
{
    int i;

    for (i = 0; i < (sizeof(stack_instructions) / sizeof(stack_instructions[0])); i++) {
	emit_procheader(stack_instructions[i].opcode);
	emit("    /* %s */\n", stack_instructions[i].instruction);
	if (stack_instructions[i].is_push) {
	    emit("    PUSH_BYTE(context, %s);\n", stack_instructions[i].reg);
	} else {
	    emit("    %s = POP_BYTE(context);\n", stack_instructions[i].reg);
	}
	if (stack_instructions[i].eval_nz) {
	    emit("    EVAL_FLAGNZ(context, %s);\n", stack_instructions[i].reg);
	}
	emit("    context->cycles_left -= 3;\n");
	emit_proctrailer();
    }
}

struct {
    u8 opcode;
    char *instruction;
    char *reg1;
    char *reg2;
    int check_flags;
} reg_transfer_instructions[] = {
    {0xaa, "TAX", "context->reg_a", "context->reg_x", 1},
    {0xa8, "TAY", "context->reg_a", "context->reg_y", 1},
    {0xba, "TSX", "context->reg_s", "context->reg_x", 1},
    {0x8a, "TXA", "context->reg_x", "context->reg_a", 1},
    {0x9a, "TXS", "context->reg_x", "context->reg_s", 0},
    {0x98, "TYA", "context->reg_y", "context->reg_a", 1},
};

void emit_reg_transfer_instructions(void)
{
    int i;

    for (i = 0; i < (sizeof(reg_transfer_instructions) / sizeof(reg_transfer_instructions[0])); i++) {
	emit_procheader(reg_transfer_instructions[i].opcode);
	emit("    /* %s */\n", reg_transfer_instructions[i].instruction);
	emit("    %s = %s;\n", reg_transfer_instructions[i].reg2, reg_transfer_instructions[i].reg1);
	emit("    context->cycles_left -= 2;\n");
	if (reg_transfer_instructions[i].check_flags) {
	    emit("    EVAL_FLAGNZ(context, %s);\n", reg_transfer_instructions[i].reg2);
	}
	emit_proctrailer();
    }
}

struct {
    u8 opcode;
    char *instruction;
    struct amode *amode;
} bit_test_instructions[] = {
    {0x24, "BIT", &amode_zpage},
    {0x2c, "BIT", &amode_abs},
};

void emit_bit_test_instructions(void)
{
    int i;
    struct amode *cur_amode;

    for (i = 0; i < (sizeof(bit_test_instructions) / sizeof(bit_test_instructions[0])); i++) {
	cur_amode = bit_test_instructions[i].amode;
    
	emit_procheader(bit_test_instructions[i].opcode);
	emit("    /* %s %s */\n", bit_test_instructions[i].instruction, cur_amode->admode);
	emit("    u8 data;\n");
	emit(cur_amode->preblock);
	emit(cur_amode->readblock, "data");
	emit("    context->flags &= ~(0xc0 | FLAG_Z);\n");
	emit("    context->flags |= data & 0xc0;\n");
	emit("    if (!(context->reg_a & data)) {\n");
	emit("        context->flags |= FLAG_Z;\n");
	emit("    }\n");
	emit("    context->cycles_left -= %d;\n", cur_amode->cycles);
	emit_proctrailer();
    }
}

struct amode amode_rega = {
    0, "Acc",
    "\n",
    "    %s = context->reg_a;\n",
    "    context->reg_a = %s;\n",
};

struct amode amode_regx = {
    0, "",
    "\n",
    "    %s = context->reg_x;\n",
    "    context->reg_x = %s;\n",
};

struct amode amode_regy = {
    0, "",
    "\n",
    "    %s = context->reg_y;\n",
    "    context->reg_y = %s;\n",
};

struct {
    u8 opcode;
    char *instruction;
    char operation;
    struct amode *amode;
} inc_dec_instructions[] = {
    {0xe6, "INC", '+', &amode_zpage},
    {0xf6, "INC", '+', &amode_zpagex},
    {0xee, "INC", '+', &amode_abs},
    {0xfe, "INC", '+', &amode_absx},
    {0xe8, "INX", '+', &amode_regx},
    {0xc8, "INY", '+', &amode_regy},
    {0xc6, "DEC", '-', &amode_zpage},
    {0xd6, "DEC", '-', &amode_zpagex},
    {0xce, "DEC", '-', &amode_abs},
    {0xde, "DEC", '-', &amode_absx},
    {0xca, "DEX", '-', &amode_regx},
    {0x88, "DEY", '-', &amode_regy},
};

void emit_inc_dec_instructions(void)
{
    int i;
    struct amode *cur_amode;

    for (i = 0; i < (sizeof(inc_dec_instructions) / sizeof(inc_dec_instructions[0])); i++) {
	cur_amode = inc_dec_instructions[i].amode;
    
	emit_procheader(inc_dec_instructions[i].opcode);
	emit("    /* %s %s */\n", inc_dec_instructions[i].instruction, cur_amode->admode);
	emit("    u8 data;\n");
	emit(cur_amode->preblock);
	emit(cur_amode->readblock, "data");
	emit("    data %c= 1;\n", inc_dec_instructions[i].operation);
	emit(cur_amode->writeblock, "data");
	emit("    EVAL_FLAGNZ(context, data);\n");
	emit("    context->cycles_left -= %d;\n", cur_amode->cycles + 2);
	emit_proctrailer();
    }
}

struct {
    u8 opcode;
    char *instruction;
    int is_store;
    char *reg;
    struct amode *amode;
} load_store_instructions[] = {
    {0xa2, "LDX", 0, "context->reg_x", &amode_immediate},
    {0xa6, "LDX", 0, "context->reg_x", &amode_zpage},
    {0xb6, "LDX", 0, "context->reg_x", &amode_zpagey},
    {0xae, "LDX", 0, "context->reg_x", &amode_abs},
    {0xbe, "LDX", 0, "context->reg_x", &amode_absy},
    {0xa0, "LDY", 0, "context->reg_y", &amode_immediate},
    {0xa4, "LDY", 0, "context->reg_y", &amode_zpage},
    {0xb4, "LDY", 0, "context->reg_y", &amode_zpagex},
    {0xac, "LDY", 0, "context->reg_y", &amode_abs},
    {0xbc, "LDY", 0, "context->reg_y", &amode_absx},
    {0x86, "STX", 1, "context->reg_x", &amode_zpage},
    {0x96, "STX", 1, "context->reg_x", &amode_zpagey},
    {0x8e, "STX", 1, "context->reg_x", &amode_abs},
    {0x84, "STY", 1, "context->reg_y", &amode_zpage},
    {0x94, "STY", 1, "context->reg_y", &amode_zpagex},
    {0x8c, "STY", 1, "context->reg_y", &amode_abs},
};

void emit_load_store_instructions(void)
{
    int i;
    struct amode *cur_amode;

    for (i = 0; i < (sizeof(load_store_instructions) / sizeof(load_store_instructions[0])); i++) {
	cur_amode = load_store_instructions[i].amode;
    
	emit_procheader(load_store_instructions[i].opcode);
	emit("    /* %s %s */\n", load_store_instructions[i].instruction, cur_amode->admode);
	emit(cur_amode->preblock);
	if (!(load_store_instructions[i].is_store)) {
	    emit(cur_amode->readblock, load_store_instructions[i].reg);
	    emit("    EVAL_FLAGNZ(context, %s);\n", load_store_instructions[i].reg);
	} else {
	    emit(cur_amode->writeblock, load_store_instructions[i].reg);
	}
	emit("    context->cycles_left -= %d;\n", cur_amode->cycles);
	emit_proctrailer();
    }
}

struct shift_rotate_instruction {
    u8 opcode;
    char *instruction;
    int is_rotate;
    int is_left;
} shift_rotate_instructions[] = {
    {0x00, "ASL", 0, 1},
    {0x20, "ROL", 1, 1},
    {0x40, "LSR", 0, 0},
    {0x60, "ROR", 1, 0},
    {0x00, NULL, 0, 0}, /* must be last entry */
};

struct shift_rotate_amode {
    u8 opcode_delta;
    struct amode *amode;
} shift_rotate_amodes[] = {
    {0x0a, &amode_rega},
    {0x06, &amode_zpage},
    {0x16, &amode_zpagex},
    {0x0e, &amode_abs},
    {0x1e, &amode_absx},
    {0x00, NULL}, /* must be last entry */
};

void emit_shift_rotate_instructions(void)
{
    struct shift_rotate_instruction *cur_instr;
    struct shift_rotate_amode *cur_amode;
    
    for (cur_instr = shift_rotate_instructions; cur_instr->instruction; cur_instr++) {
	for (cur_amode = shift_rotate_amodes; cur_amode->amode; cur_amode++) {
	    emit_procheader(cur_instr->opcode + cur_amode->opcode_delta);
	    emit("    /* %s %s */\n", cur_instr->instruction, cur_amode->amode->admode);
	    emit("    u8 src_data;\n");
	    emit("    u8 dst_data;\n");
	    emit(cur_amode->amode->preblock);
	    emit(cur_amode->amode->readblock, "src_data");
	    emit("    dst_data = src_data %s 1;\n", cur_instr->is_left? "<<": ">>");
	    if (cur_instr->is_rotate) {
		if (cur_instr->is_left) {
		    emit("    dst_data |= context->flags & FLAG_C;\n");
		} else {
		    emit("    dst_data |= context->flags << 7;\n");
		}
	    }
	    emit(cur_amode->amode->writeblock, "dst_data");
	    emit("    context->flags &= ~FLAG_C;\n");
	    if (cur_instr->is_left) {
		emit("    context->flags |= src_data >> 7;\n");
	    } else {
		emit("    context->flags |= src_data & FLAG_C;\n");
	    }
	    emit("    EVAL_FLAGNZ(context, dst_data);\n");
	    emit("    context->cycles_left -= %d;\n", cur_amode->amode->cycles + 2);
	    emit_proctrailer();
	}
    }
}

struct compare_amode {
    u8 opcode_delta;
    struct amode *amode;
} compare_amodes[] = {
    {0xc0, &amode_immediate},
    {0xc4, &amode_zpage},
    {0xcc, &amode_abs},
    {0x00, NULL},
};

void emit_compare_instructions(void)
{
    int i;
    struct compare_amode *cur_amode;

    for (i = 0; i < 2; i++) {
	for (cur_amode = compare_amodes; cur_amode->amode; cur_amode++) {
	    emit_procheader((i << 5) + cur_amode->opcode_delta);
	    emit("    /* CP%c %s */\n", 'Y' - i, cur_amode->amode->admode);
	    emit("    u8 tmp;\n");
	    emit("    u8 result;\n");
	    
	    emit(cur_amode->amode->preblock);
	    emit(cur_amode->amode->readblock, "tmp");
	    
	    emit("    result = context->reg_%c - tmp;\n", 'y' - i);
	    emit("    if (context->reg_%c < tmp) {\n", 'y' - i);
	    emit("        context->flags &= ~FLAG_C;\n");
	    emit("    } else {\n");
	    emit("        context->flags |= FLAG_C;\n");
	    emit("    }\n");
	    emit("    EVAL_FLAGNZ(context, result);\n");
	    emit("    context->cycles_left -= %d;\n", cur_amode->amode->cycles);
	    emit_proctrailer();
	}
    }
}

struct misc_instruction {
    u8 opcode;
    char *instruction;
} misc_instructions[] = {
    /* call/return group */
    {0x4c, "    /* JMP Abs */\n"
     "    u16 addr;\n"
     "\n"
     "    addr = OPFETCH(context);\n"
     "    addr |= OPFETCH(context) << 8;\n"
     "    context->pc = addr;\n"
     "    context->cycles_left -= 4;\n"},
    {0x6c, "    /* JMP Ind */\n"
     "    u16 addr;\n"
     "\n"
     "    addr = OPFETCH(context);\n"
     "    addr |= OPFETCH(context) << 8;\n"
     "    context->pc = context->readfunc(context->cpu, addr);\n"
     "    addr = (addr & 0xff00) | ((addr + 1) & 0x00ff);\n"
     "    context->pc |= context->readfunc(context->cpu, addr) << 8;\n"
     "    context->cycles_left -= 7;\n"},
    {0x20, "    /* JSR */\n"
     "    u16 addr;\n"
     "\n"
     "    addr = OPFETCH(context);\n"
     "    PUSH_BYTE(context, context->pc >> 8);\n"
     "    PUSH_BYTE(context, context->pc & 0xff);\n"
     "    addr |= OPFETCH(context) << 8;\n"
     "    context->pc = addr;\n"
     "    context->cycles_left -= 7;\n"},
    {0x60, "    /* RTS */\n"
     "\n"
     "    context->pc = POP_BYTE(context);\n"
     "    context->pc |= POP_BYTE(context) << 8;\n"
     "    context->pc++;\n"
     "    context->cycles_left -= 6;\n"},
    {0x40, "    /* RTI */\n"
     "\n"
     "    context->flags = POP_BYTE(context);\n"
     "    context->pc = POP_BYTE(context);\n"
     "    context->pc |= POP_BYTE(context) << 8;\n"
     "    context->cycles_left -= 6;\n"},
    {0x00, "    /* BRK */\n"
     "\n"
     "    context->pc++;\n"
     "    PUSH_BYTE(context, context->pc >> 8);\n"
     "    PUSH_BYTE(context, context->pc & 0xff);\n"
     "    PUSH_BYTE(context, context->flags | FLAG_B);\n"
     "    context->flags |= FLAG_I;\n"
     "    context->flags &= ~FLAG_D;\n"
     "    context->pc = context->readfunc(context->cpu, VECTOR_BRK_LO);\n"
     "    context->pc |= context->readfunc(context->cpu, VECTOR_BRK_HI) << 8;\n"
     "    context->cycles_left -= 7;\n"},
    {0xea, "    /* NOP */\n"},
    {0x00, NULL}, /* must be last entry */
};

void emit_misc_instructions(void)
{
    struct misc_instruction *cur_instr;

    for (cur_instr = misc_instructions; cur_instr->instruction; cur_instr++) {
	emit_procheader(cur_instr->opcode);
	emit(cur_instr->instruction);
	emit_proctrailer();
    }
}

instr_emitter_t emitter_list[] = {
    emit_flag_instructions,
    emit_big8_instructions,
    emit_branch_instructions,
    emit_stack_instructions,
    emit_reg_transfer_instructions,
    emit_bit_test_instructions,
    emit_inc_dec_instructions,
    emit_load_store_instructions,
    emit_shift_rotate_instructions,
    emit_compare_instructions,
    emit_misc_instructions,
    NULL, /* must be last entry */
};

void build_core(void)
{
    int i;

    for (i = 0; emitter_list[i]; i++) {
	emitter_list[i]();
    }
}

int main(int argc, char *argv[])
{
    if (argc != 2) {
	printf("usage: %s <outfile>\n", argv[0]);
	return 1;
    }

    outfile = fopen(argv[1], "w");
    if (!outfile) {
	perror("Unable to open file for writing");
	return 1;
    }

    emit_file_header(argv[1]);

    clear_handlers();

    build_core();
    
    emit_decode_table();
    
    emit_file_footer();

    if (fclose(outfile) == EOF) {
	perror("Closing file");
	return 1;
    }
    
    return 0;
}

/*
 * $Log: build_cpu6502.c,v $
 * Revision 1.1  2000/05/01 00:34:17  nyef
 * Initial revision
 *
 */
